"""Test OpenAPI handling."""

from __future__ import annotations

import typing as t

import pytest

from tap_shortcut.tap import handle_x_nullable


@pytest.mark.parametrize(
    ("input_schema", "expected"),
    [
        (
            {
                "type": "object",
                "required": ["id"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": "string",
                        "x-nullable": True,
                    },
                },
            },
            {
                "type": "object",
                "required": ["id"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": ["string", "null"],
                        "x-nullable": True,
                    },
                },
            },
        ),
        (
            {
                "type": "object",
                "required": ["id"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": "string",
                        "x-nullable": True,
                    },
                    "object_prop": {
                        "type": "object",
                        "properties": {
                            "nested": {
                                "type": "string",
                                "x-nullable": True,
                            },
                        },
                        "x-nullable": True,
                    },
                },
            },
            {
                "type": "object",
                "required": ["id"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": ["string", "null"],
                        "x-nullable": True,
                    },
                    "object_prop": {
                        "type": ["object", "null"],
                        "properties": {
                            "nested": {
                                "type": ["string", "null"],
                                "x-nullable": True,
                            },
                        },
                        "x-nullable": True,
                    },
                },
            },
        ),
        (
            {
                "type": "object",
                "required": ["id", "array_prop"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": "string",
                        "x-nullable": True,
                    },
                    "array_prop": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "required": ["id"],
                            "properties": {
                                "id": {
                                    "type": "integer",
                                },
                                "nested": {
                                    "type": "string",
                                    "x-nullable": True,
                                },
                            },
                        },
                    },
                },
            },
            {
                "type": "object",
                "required": ["id", "array_prop"],
                "properties": {
                    "id": {
                        "type": "integer",
                    },
                    "name": {
                        "type": ["string", "null"],
                        "x-nullable": True,
                    },
                    "array_prop": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "required": ["id"],
                            "properties": {
                                "id": {
                                    "type": "integer",
                                },
                                "nested": {
                                    "type": ["string", "null"],
                                    "x-nullable": True,
                                },
                            },
                        },
                    },
                },
            },
        ),
    ],
    ids=["simple_object", "nested_object", "nested_array"],
)
def test_handle_x_nullable(
    input_schema: dict[str, t.Any],
    expected: dict[str, t.Any],
) -> None:
    """Test handling of x-nullable."""
    assert handle_x_nullable(input_schema) == expected



"""Shortcut tap class."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

import requests
from singer_sdk import RESTStream, Stream, Tap
from singer_sdk import typing as th
from singer_sdk._singerlib import resolve_schema_references
from toolz.dicttoolz import get_in

from tap_shortcut import streams

if TYPE_CHECKING:
    from tap_shortcut.client import ShortcutStream

STREAM_TYPES: list[type[ShortcutStream]] = [
    streams.Categories,
    streams.Epics,
    streams.Files,
    streams.Groups,
    streams.Iterations,
    streams.Labels,
    streams.Members,
    streams.Milestones,
    streams.Projects,
    streams.ProjectStories,
    streams.Repositories,
    streams.Workflows,
]

OPENAPI_URL = "https://developer.shortcut.com/api/rest/v3/shortcut.swagger.json"


def handle_x_nullable(schema: dict[str, Any]) -> dict[str, Any]:
    """Resolve x-nullable properties to standard JSON Schema nullable type.

    Args:
        schema: A JSON Schema dictionary.

    Returns:
        A new JSON Schema dictionary with 'x-nullable' resolved to [<type>, "null"].
    """
    result = deepcopy(schema)

    if "object" in result["type"]:
        for prop, prop_schema in result.get("properties", {}).items():
            prop_type: str | list[str] = prop_schema.get("type", [])
            types = [prop_type] if isinstance(prop_type, str) else prop_type
            nullable: bool = prop_schema.get("x-nullable", False)

            if nullable:
                prop_schema["type"] = [*types, "null"]

            result["properties"][prop] = handle_x_nullable(prop_schema)

    elif "array" in result["type"]:
        result["items"] = handle_x_nullable(result["items"])

    if "enum" in result and None not in result["enum"]:
        result["enum"].append(None)

    return result


class TapShortcut(Tap):
    """Singer tap for Shortcut."""

    name = "tap-shortcut"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "token",
            th.StringType,
            required=True,
            description="Shortcut Token",
        ),
    ).to_dict()

    def get_openapi_schema(self) -> dict[Any, Any]:
        """Retrieve Swagger/OpenAPI schema for this API.

        Returns:
            OpenAPI schema.

        Raises:
            RuntimeError: If the OpenAPI schema cannot be retrieved.
        """
        response = requests.get(OPENAPI_URL, timeout=30)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            msg = f"Error retrieving OpenAPI schema ({err})"
            raise RuntimeError(msg) from err

        return response.json()  # type: ignore[no-any-return]

    def discover_streams(self) -> list[Stream]:
        """Return a list of discovered streams.

        Returns:
            A list of Shortcut streams.
        """
        openapi_schema = self.get_openapi_schema()
        streams: list[RESTStream[None]] = []

        for stream_type in STREAM_TYPES:
            schema = get_in(
                keys=[
                    "paths",
                    stream_type.path,
                    "get",
                    "responses",
                    "200",
                    "schema",
                    "items",
                ],
                coll=openapi_schema,
            )
            schema["definitions"] = openapi_schema["definitions"]
            resolved_schema = resolve_schema_references(schema)
            clean_schema = handle_x_nullable(resolved_schema)
            stream_type.preprocess_schema(clean_schema)

            streams.append(stream_type(tap=self, schema=clean_schema))

        return sorted(streams, key=lambda x: x.name)



"""Stream type classes for tap-shortcut."""

from __future__ import annotations

import typing as t

from tap_shortcut.client import ShortcutStream

__all__ = [
    "Categories",
    "Epics",
    "Files",
    "Groups",
    "Iterations",
    "Labels",
    "Members",
    "Milestones",
    "Projects",
    "ProjectStories",
    "Repositories",
    "Workflows",
]


class Members(ShortcutStream):
    """Members stream."""

    name = "members"
    path = "/api/v3/members"
    extra_nullable_fields = ("replaced_by",)


class Projects(ShortcutStream):
    """Projects stream."""

    name = "projects"
    path = "/api/v3/projects"

    def get_child_context(
        self,
        record: dict[str, t.Any],
        context: dict[str, t.Any] | None,  # noqa: ARG002
    ) -> dict[str, t.Any]:
        """Return a dictionary of child context.

        Args:
            record: A dictionary of the record.
            context: A dictionary of the context.

        Returns:
            A dictionary of the child context.
        """
        return {"project-public-id": record["id"]}


class ProjectStories(ShortcutStream):
    """Project stories stream."""

    name = "project_stories"
    path = "/api/v3/projects/{project-public-id}/stories"
    parent_stream_type = Projects
    ignore_parent_replication_key = True
    extra_nullable_fields = (
        "description",
        "synced_item",
        "unresolved_blocker_comments",
        "lead_time",
        "cycle_time",
    )

    @classmethod
    def preprocess_schema(cls: type[ProjectStories], schema: dict[str, t.Any]) -> None:
        """Return the schema of the stream."""
        super().preprocess_schema(schema)
        schema["properties"]["lead_time"]["type"] = ["number", "null"]
        schema["properties"]["cycle_time"]["type"] = ["number", "null"]


class Epics(ShortcutStream):
    """Epics stream."""

    name = "epics"
    path = "/api/v3/epics"
    extra_nullable_fields = ("description",)


class Workflows(ShortcutStream):
    """Workflows stream."""

    name = "workflows"
    path = "/api/v3/workflows"


class Milestones(ShortcutStream):
    """Milestones stream."""

    name = "milestones"
    path = "/api/v3/milestones"


class Labels(ShortcutStream):
    """Labels stream."""

    name = "labels"
    path = "/api/v3/labels"


class Categories(ShortcutStream):
    """Categories stream."""

    name = "categories"
    path = "/api/v3/categories"


class Files(ShortcutStream):
    """Files stream."""

    name = "files"
    path = "/api/v3/files"


class Groups(ShortcutStream):
    """Groups stream."""

    name = "groups"
    path = "/api/v3/groups"


class Iterations(ShortcutStream):
    """Iterations stream."""

    name = "iterations"
    path = "/api/v3/iterations"


class Repositories(ShortcutStream):
    """Repositories stream."""

    name = "repositories"
    path = "/api/v3/repositories"



"""Tests standard tap features using the built-in SDK tests library."""

from __future__ import annotations

from typing import Any

from singer_sdk.testing import SuiteConfig, get_tap_test_class

from tap_shortcut.tap import TapShortcut

SAMPLE_CONFIG: dict[str, Any] = {}


TestTapShortcut = get_tap_test_class(
    TapShortcut,
    config=SAMPLE_CONFIG,
    suite_config=SuiteConfig(
        ignore_no_records_for_streams=[
            "categories",
            "files",
            "iterations",
            "milestones",
        ]
    ),
)



"""REST client handling, including ShortcutStream base class."""

from __future__ import annotations

import typing as t

from singer_sdk import RESTStream
from singer_sdk.authenticators import APIKeyAuthenticator

T = t.TypeVar("T", bound="ShortcutStream")


def _handle_runtime_nullables(
    schema: dict[str, t.Any],
    fields: t.Iterable[str],
) -> None:
    """Force fields to be nullable at runtime."""
    for field in fields:
        if field in schema["properties"]:
            prop_type: str | list[str] = schema["type"]
            types = [prop_type] if isinstance(prop_type, str) else prop_type
            schema["properties"][field]["type"] = [*types, "null"]


class ShortcutStream(RESTStream[None]):
    """Shortcut stream class."""

    url_base = "https://api.app.shortcut.com"
    records_jsonpath = "$[*]"
    primary_keys: t.ClassVar[list[str]] = ["id"]

    # Abstract fields
    extra_nullable_fields: tuple[str, ...] = ()

    @property
    def authenticator(self) -> APIKeyAuthenticator:
        """Get an authenticator object.

        Returns:
            The authenticator instance for this REST stream.
        """
        token: str = self.config["token"]
        return APIKeyAuthenticator.create_for_stream(
            self,
            key="Shortcut-Token",
            value=token,
            location="header",
        )

    @property
    def http_headers(self) -> dict[str, str]:
        """Return the http headers needed.

        Returns:
            A dictionary of HTTP headers.
        """
        return {"User-Agent": f"{self.tap_name}/{self._tap.plugin_version}"}

    @classmethod
    def preprocess_schema(cls: type[T], schema: dict[str, t.Any]) -> None:
        """Preprocess the schema.

        Args:
            schema: A dictionary of the schema.

        Returns:
            A dictionary of the processed schema.
        """
        _handle_runtime_nullables(schema, cls.extra_nullable_fields)
